"""
Miscellaneous functions used by Topographica code outside of topo.base.

Several of these files are completely self-contained, not importing
anything from Topographica, and can thus be copied and used in
unrelated projects.

$Id$
"""

# For backwards compatibility; this file used to be in misc/
from imagen import patternfn # pyflakes:ignore (API import)


